"""
This package contains functions to enhance and/or detect objects of different
shapes (e.g. blobs, vessels)
"""

# make functions available at the package level using shadow imports
# since we mostly have one function per file
from .cdog import cdog
from .clog import clog
from .glog import glog
from .vesselness import vesselness

# list out things that are available for public use
__all__ = (

    # functions and classes of this package
    'cdog',
    'clog',
    'glog',
    'vesselness',
)



import numpy as np


def delete(im_label, indices):
    """
    Deletes objects with values in 'indices' from label image, writing them over
    with zeros to assimilate with background.

    Parameters
    ----------
    im_label : array_like
        A label image generated by segmentation methods.
    indices : array_like
        An n-length array of strictly positive integer values to delete from
        'im_label'.

    Returns
    -------
    Deleted : array_like
        A label image where all values in 'indices' are set to zero.

    Notes
    -----
    A call to CondenseLabel can squeeze label image values to fill in gaps from
    deleted values.

    See Also
    --------
    histomicstk.segmentation.label.condense

    """
    import scipy.ndimage as ndi

    # initialize output
    Deleted = im_label.copy()

    # get extent of each object
    Locations = ndi.find_objects(Deleted)

    # fill in new values
    for i in np.arange(indices.size):
        if Locations[indices[i] - 1] is not None:
            Patch = Deleted[Locations[indices[i] - 1]]
            Patch[Patch == indices[i]] = 0

    return Deleted



import json
import pprint
import time
from pathlib import Path

import large_image
import numpy as np

import histomicstk
import histomicstk.segmentation.positive_pixel_count as ppc
from histomicstk.cli import utils
from histomicstk.cli.utils import CLIArgumentParser


def tile_positive_pixel_count(imagePath, tilePosition, it_kwargs, ppc_params,
                              color_map, useAlpha, region_polygons, style):
    # Start measuring time
    tile_start_time = time.time()

    # Get the tile source
    ts = large_image.getTileSource(imagePath, style=style)

    # Fetch the image tile
    tile = ts.getSingleTile(tile_position=tilePosition, **it_kwargs)

    # Convert region polygons to binary mask
    mask = utils.polygons_to_binary_mask(
        region_polygons, tile['x'], tile['y'], tile['width'], tile['height'])

    # Get the tile image
    img = tile['tile']

    # Handle single-channel images (convert to 3 channels)
    if len(img.shape) == 2:
        img = img.reshape((img.shape[0], img.shape[1], 1))
    if len(img.shape) == 3 and img.shape[-1] == 1:
        img = np.repeat(img, 3, axis=2)

    # Calculate the positive pixel count and generate a mask
    result, ppcmask = ppc.count_image(img, ppc_params, mask)

    # Release the tile to free up resources
    tile.release()

    # Apply the color map to the mask to generate an image with colored regions
    ppcimg = color_map[ppcmask]

    # Optionally remove alpha channel if not used
    if not useAlpha:
        ppcimg = ppcimg[:, :, :3]

    # Return the result, colored image, tile's x and y coordinates, mask, and
    # tile processing start time
    return result, ppcimg, tile['x'], tile['y'], mask, tile_start_time


def main(opts):
    # Print the options dictionary using pprint
    pprint.pprint(vars(opts))

    # Handle style option
    if not opts.style or opts.style.startswith('{#control'):
        opts.style = None

    # Get the tile source
    ts = large_image.getTileSource(opts.inputImageFile, style=opts.style)

    # Create a sink for the output label image if specified
    sink = large_image.new() if getattr(opts, 'outputLabelImage', None) else None

    # Get region parameters and polygons
    tiparams = utils.get_region_dict(opts.region, None, ts)
    region_polygons = utils.get_region_polygons(opts.region)

    # Print region parameters and polygons
    print('region: %r %r' % (tiparams, region_polygons))

    # Set tile size and useAlpha flag
    tileSize = 4096
    useAlpha = len(opts.region) > 6

    # Define the color map for label images
    color_map = np.empty((4, 4), dtype=np.uint8)
    color_map[ppc.Labels.NEGATIVE] = 255, 255, 255, 255
    color_map[ppc.Labels.WEAK] = 60, 78, 194, 255
    color_map[ppc.Labels.PLAIN] = 221, 220, 220, 255
    color_map[ppc.Labels.STRONG] = 180, 4, 38, 255

    # Set positive pixel count parameters
    ppc_params = ppc.Parameters(
        **{k: getattr(opts, k) for k in ppc.Parameters._fields},
    )

    # Initialize a list to store results
    results = []

    # Set crop coordinates if 'left' is present in the region parameters
    if sink and 'left' in tiparams.get('region', {}):
        sink.crop = (
            tiparams['region']['left'], tiparams['region']['top'],
            tiparams['region']['width'], tiparams['region']['height'])

    # Set parameters for tile source fetching
    tiparams['format'] = large_image.constants.TILE_FORMAT_NUMPY
    tiparams['tile_size'] = dict(width=tileSize, height=tileSize)
    try:
        tiparams['frame'] = int(opts.frame)
    except Exception:
        pass

    # Get the total number of tiles
    tileCount = next(ts.tileIterator(**tiparams))['iterator_range']['position']

    # Start processing tiles
    start_time = time.time()

    if tileCount > 4 and getattr(opts, 'scheduler', None) != 'none':
        # If tile count is large and Dask scheduler is not disabled, use Dask for
        # parallel processing
        print('>> Creating Dask client')
        client = utils.create_dask_client(opts)
        dask_setup_time = time.time() - start_time
        print(f'Dask setup time = {utils.disp_time_hms(dask_setup_time)}')
        futureList = []
        for tile in ts.tileIterator(**tiparams):
            tile_position = tile['tile_position']['position']
            # Submit the tile processing function to Dask client
            futureList.append(client.submit(
                tile_positive_pixel_count,
                opts.inputImageFile, tile_position, tiparams, ppc_params,
                color_map, useAlpha, region_polygons, opts.style))
        for idx, future in enumerate(futureList):
            # Wait for the results from Dask futures and process each tile
            result, ppcimg, x, y, mask, tile_start_time = future.result()
            results.append(result)
            if sink:
                sink.addTile(ppcimg, x, y, mask=mask)
            print('Processed tile %d/%d\n  %r\n  time %s (%s from start)' % (
                idx, tileCount, result,
                utils.disp_time_hms(time.time() - tile_start_time),
                utils.disp_time_hms(time.time() - start_time)))
    else:
        # If tile count is small or Dask scheduler is disabled, process tiles sequentially
        for tile in ts.tileIterator(**tiparams):
            tile_position = tile['tile_position']['position']
            result, ppcimg, x, y, mask, tile_start_time = tile_positive_pixel_count(
                opts.inputImageFile, tile_position, tiparams, ppc_params,
                color_map, useAlpha, region_polygons, opts.style)
            results.append(result)
            if sink:
                sink.addTile(ppcimg, x, y, mask=mask)
            print('Processed tile %d/%d\n  %r\n  time %s (%s from start)' % (
                tile_position, tileCount, result,
                utils.disp_time_hms(time.time() - tile_start_time),
                utils.disp_time_hms(time.time() - start_time)))

    print('Combining results, time from start %s' % (utils.disp_time_hms(time.time() - start_time)))

    # Combine the results and calculate statistics
    stats = ppc._totals_to_stats(ppc._combine(results))

    if sink:
        # Output the result to the label image file if specified
        print('Outputting file, time from start %s' % (
            utils.disp_time_hms(time.time() - start_time)))
        sink_start_time = time.time()
        sink.write(opts.outputLabelImage, lossy=False)
        print('Output file, time %s (%s from start)' % (
            utils.disp_time_hms(time.time() - sink_start_time),
            utils.disp_time_hms(time.time() - start_time)))

    # Print the calculated statistics
    pprint.pprint(stats._asdict())

    # Create an annotation dictionary
    annotation = {
        'name': 'Positive Pixel Count',
        'elements': [{
            'type': 'image',
            'girderId': 'outputLabelImage',
            'hasAlpha': useAlpha,
            'transform': {
                'xoffset': tiparams.get('region', {}).get('left', 0),
                'yoffset': tiparams.get('region', {}).get('top', 0),
            },
        }],
        'attributes': {
            'params': vars(opts),
            'stats': stats._asdict(),
            'cli': Path(__file__).stem,
            'version': histomicstk.__version__,
        },
    }

    # Save the annotation dictionary to the output annotation file
    if opts.outputAnnotationFile:
        with open(opts.outputAnnotationFile, 'w') as annotation_file:
            json.dump(annotation, annotation_file, separators=(',', ':'), sort_keys=False)
        print('Finished time %s' % (utils.disp_time_hms(time.time() - start_time)))


if __name__ == '__main__':
    main(CLIArgumentParser().parse_args())



"""Placeholder."""
import numpy

import histomicstk.utils as utils

from . import _linalg as linalg
from .complement_stain_matrix import complement_stain_matrix


def separate_stains_macenko_pca(
        im_sda, minimum_magnitude=16, min_angle_percentile=0.01,
        max_angle_percentile=0.99, mask_out=None):
    """Compute the stain matrix for color deconvolution with the Macenko method.

    For a two-stain image or matrix in SDA space, this method works by
    computing a best-fit plane with PCA, wherein it selects the stain
    vectors as percentiles in the "angle distribution" in that plane.

    Parameters
    ----------
    im_sda : array_like
        Image (MxNx3) or matrix (3xN) in SDA space for which to compute the
        stain matrix.
    minimum_magnitude : float
        The magnitude below which vectors will be excluded from the computation
        of the angle distribution.

        The default is based on the paper value of 0.15, adjusted for our
        method of calculating SDA, thus 0.15 * 255 * log(10)/log(255)
    min_angle_percentile : float
        The smaller percentile of one of the vectors to pick from the angle
        distribution
    max_angle_percentile : float
        The larger percentile of one of the vectors to pick from the angle
        distribution
    mask_out : array_like
        if not None, should be (m, n) boolean numpy array.
        This parameter ensures exclusion of non-masked areas from calculations.
        This is relevant because elements like blood, sharpie marker,
        white space, etc may throw off the normalization somewhat.

    Returns
    -------
    w : array_like
        A 3x3 matrix of stain column vectors

    Note
    ----
    All input pixels not otherwise excluded are used in the computation of the
    principal plane and the angle distribution.

    See Also
    --------
    histomicstk.preprocessing.color_deconvolution.color_deconvolution
    histomicstk.preprocessing.color_deconvolution.separate_stains_xu_snmf

    References
    ----------
    .. [#] Van Eycke, Y. R., Allard, J., Salmon, I., Debeir, O., &
           Decaestecker, C. (2017).  Image processing in digital pathology: an
           opportunity to solve inter-batch variability of immunohistochemical
           staining.  Scientific Reports, 7.
    .. [#] Macenko, M., Niethammer, M., Marron, J. S., Borland, D.,
           Woosley, J. T., Guan, X., ... & Thomas, N. E. (2009, June).
           A method for normalizing histology slides for quantitative analysis.
           In Biomedical Imaging: From Nano to Macro, 2009.  ISBI'09.
           IEEE International Symposium on (pp. 1107-1110). IEEE.

    """
    # Image matrix
    m = utils.convert_image_to_matrix(im_sda)

    # mask out irrelevant values
    if mask_out is not None:
        keep_mask = numpy.equal(mask_out[..., None], False)
        keep_mask = numpy.tile(keep_mask, (1, 1, 3))
        keep_mask = utils.convert_image_to_matrix(keep_mask)
        m = m[:, keep_mask.all(axis=0)]

    # get rid of NANs and infinities
    m = utils.exclude_nonfinite(m)

    # Principal components matrix
    pcs = linalg.get_principal_components(m)
    # Input pixels projected into the PCA plane
    proj = pcs.T[:-1].dot(m)
    # Pixels above the magnitude threshold
    filt = proj[:, linalg.magnitude(proj) > minimum_magnitude]
    # The "angles"
    angles = _get_angles(filt)

    # The stain vectors

    def get_percentile_vector(p):
        return pcs[:, :-1].dot(filt[:, argpercentile(angles, p)])

    min_v = get_percentile_vector(min_angle_percentile)
    max_v = get_percentile_vector(max_angle_percentile)

    # The stain matrix
    w = complement_stain_matrix(linalg.normalize(
        numpy.array([min_v, max_v]).T))
    return w


def _get_angles(m):
    """Take a 2xN matrix of vectors and return a length-N array of an.

    ... angle-like quantity.
    Since this is an internal function, we assume that the values
    result from PCA, and so the second element of the vectors captures
    secondary variation -- and thus is the one that takes on both
    positive and negative values.

    """
    m = linalg.normalize(m)
    # "Angle" towards +x from the +y axis
    return (1 - m[1]) * numpy.sign(m[0])


def argpercentile(arr, p):
    """Calculate index in arr of element nearest the pth percentile."""
    # Index corresponding to percentile
    i = min(int(p * arr.size + 0.5), arr.size - 1)
    return numpy.argpartition(arr, i)[i]



import warnings

import numpy as np


def compute_morphometry_features(im_label, rprops=None):
    """
    Calculate morphometry features for each object

    Parameters
    ----------
    im_label : array_like
        A labeled mask image wherein intensity of a pixel is the ID of the
        object it belongs to. Non-zero values are considered to be foreground
        objects.

    rprops : output of skimage.measure.regionprops, optional
        rprops = skimage.measure.regionprops( im_label ). If rprops is not
        passed then it will be computed inside which will increase the
        computation time.

    Returns
    -------
    fdata: pandas.DataFrame
        A pandas dataframe containing the morphometry features for each
        object/label listed below.

    Notes
    -----
    List of morphometry features computed by this function:

    Orientation.Orientation :  float
        Angle between the horizontal axis and the major axis of the ellipse
        that has the same second moments as the region,
        ranging from `-pi/2` to `pi/2` counter-clockwise.

    Size.Area : int
        Number of pixels the object occupies.

    Size.ConvexHullArea :  int
        Number of pixels of convex hull image, which is the smallest convex
        polygon that encloses the region.

    Size.MajorAxisLength : float
        The length of the major axis of the ellipse that has the same
        normalized second central moments as the object.

    Size.MinorAxisLength : float
        The length of the minor axis of the ellipse that has the same
        normalized second central moments as the region.

    Size.Perimeter : float
        Perimeter of object which approximates the contour as a line
        through the centers of border pixels using a 4-connectivity.

    Shape.Circularity: float
        A measure of how similar the shape of an object is to the circle

    Shape.Eccentricity : float
        A measure of aspect ratio computed to be the eccentricity of the
        ellipse that has the same second-moments as the object region.
        Eccentricity of an ellipse is the ratio of the focal distance
        (distance between focal points) over the major axis length. The value
        is in the interval [0, 1). When it is 0, the ellipse becomes a circle.

    Shape.EquivalentDiameter : float
        The diameter of a circle with the same area as the object.

    Shape.Extent : float
        Ratio of area of the object to its axis-aligned bounding box.

    Shape.FractalDimension : float
        Minkowski–Bouligand dimension, aka. the box-counting dimension. It
        is a measure of boundary complexity. See
        https://en.wikipedia.org/wiki/Minkowski%E2%80%93Bouligand_dimension

    Shape.MinorMajorAxisRatio : float
        A measure of aspect ratio. Ratio of minor to major axis of the ellipse
        that has the same second-moments as the object region

    Shape.Solidity : float
        A measure of convexity computed as the ratio of the number of pixels
        in the object to that of its convex hull.

    Shape.HuMoments-k : float
        Where k ranges from 1-7 are the 7 Hu moments features. The first six
        moments are translation, scale and rotation invariant, while the
        seventh moment flips its sign if the shape is a mirror image.
        See https://learnopencv.com/shape-matching-using-hu-moments-c-python/

    Shape.WeightedHuMoments-k : float
        Same as Hu moments, but instead of using the binary mask, using the
        intensity image.

    """
    import pandas as pd
    from skimage.measure import regionprops

    # compute object properties if not provided
    if rprops is None:
        rprops = regionprops(im_label)
    intensity_wtd = rprops[0]._intensity_image is not None

    # Define the feature list as tuples of (feature name, attribute or function)
    feature_list = [
        ('Orientation.Orientation', 'orientation'),
        ('Size.Area', 'area'),
        ('Size.ConvexHullArea', 'convex_area'),
        ('Size.MajorAxisLength', 'major_axis_length'),
        ('Size.MinorAxisLength', 'minor_axis_length'),
        ('Size.Perimeter', 'perimeter'),
        (
            'Shape.Circularity',
            lambda rp: 4
            * np.pi
            * rp.area
            / (rp.perimeter**2 if rp.perimeter > 0 else 1),
        ),
        ('Shape.Eccentricity', 'eccentricity'),
        ('Shape.EquivalentDiameter', 'equivalent_diameter'),
        ('Shape.Extent', 'extent'),
        ('Shape.FractalDimension', lambda rp: _fractal_dimension(rp.image)),
        (
            'Shape.MinorMajorAxisRatio',
            lambda rp: rp.minor_axis_length / rp.major_axis_length
            if rp.major_axis_length > 0
            else 1,
        ),
        ('Shape.Solidity', 'solidity'),
    ]

    # Add Hu moments features
    hu_moments = [
        ('Shape.HuMoments' + str(k), lambda rp, k=k: rp.moments_hu[k - 1])
        for k in range(1, 8)
    ]
    feature_list.extend(hu_moments)
    if intensity_wtd:
        wtd_hu_moments = [
            (
                'Shape.WeightedHuMoments' + str(k),
                lambda rp, k=k: rp.weighted_moments_hu[k - 1],
            )
            for k in range(1, 8)
        ]
        feature_list.extend(wtd_hu_moments)

    data = []

    for prop in rprops:
        row = []
        for name, attr in feature_list:
            if callable(attr):
                value = attr(prop)
            else:
                value = getattr(prop, attr, np.nan)

            # Check if the feature is Area or ConvexHullArea and ensure it is a float
            if name in ('Size.Area', 'Size.ConvexHullArea'):
                value = float(value)  # Cast to float here

            row.append(value)
        data.append(row)

    fdata = pd.DataFrame(data, columns=[name for name, _ in feature_list])

    return fdata


def _fractal_dimension(Z):
    """
    Calculate the fractal dimension of an object (boundary complexity).

    Source: https://gist.github.com/rougier/e5eafc276a4e54f516ed5559df4242c0

    From https://en.wikipedia.org/wiki/Minkowski–Bouligand_dimension ...
    In fractal geometry, the Minkowski–Bouligand dimension, also known as
    Minkowski dimension or box-counting dimension, is a way of determining the
    fractal dimension of a set S in a Euclidean space Rn, or more generally in
    a metric space (X, d).

    """
    # Only for 2d binary image
    assert len(Z.shape) == 2
    Z = Z > 0

    # From https://github.com/rougier/numpy-100 (#87)
    def boxcount(arr, k):
        S = np.add.reduceat(
            np.add.reduceat(arr, np.arange(0, arr.shape[0], k), axis=0),
            np.arange(0, arr.shape[1], k),
            axis=1)
        # We count non-empty (0) and non-full boxes (k*k)
        return len(np.where((S > 0) & (S < k * k))[0])

    # Minimal dimension of image
    p = min(Z.shape)

    # Greatest power of 2 less than or equal to p
    n = 2 ** np.floor(np.log(p) / np.log(2))

    # Extract the exponent
    n = int(np.log(n) / np.log(2))

    # Build successive box sizes (from 2**n down to 2**1)
    sizes = 2 ** np.arange(n, 1, -1)

    # Actual box counting with decreasing size
    counts = []
    for size in sizes:
        counts.append(boxcount(Z, size))

    # Fit the successive log(sizes) with log (counts)
    coeffs = [0]
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', np.RankWarning)
        if len(counts):
            try:
                coeffs = np.polyfit(np.log(sizes), np.log(counts), 1)
            except TypeError:
                pass
    return -coeffs[0]



"""
Created on Mon Aug 12 18:47:34 2019.

@author: tageldim

"""
import copy
import os
import sys

import pytest
from pandas import read_csv

from histomicstk.annotations_and_masks.annotation_and_mask_utils import (
    _get_idxs_for_all_rois, get_bboxes_from_slide_annotations,
    get_scale_factor_and_appendStr, scale_slide_annotations)
from histomicstk.annotations_and_masks.annotations_to_masks_handler import (
    get_all_rois_from_slide, get_image_and_mask_from_slide, get_roi_mask)

thisDir = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, os.path.join(thisDir, '../../../tests'))
import htk_test_utilities as utilities  # noqa
from htk_test_utilities import getTestFilePath, girderClient  # noqa


class Cfg:
    def __init__(self):
        self.gc = None
        self.GTcodes = None
        self.iteminfo = None
        self.slide_annotations = None
        self.element_infos = None
        self.get_roi_mask_kwargs = None
        self.get_contours_kwargs = None
        self.get_kwargs = None
        self.MPP = None
        self.MAG = None


cfg = Cfg()

# pytest runs tests in the order they appear in the module


@pytest.mark.usefixtures('girderClient')  # noqa
def test_prep(girderClient):  # noqa

    cfg.gc = girderClient

    original_iteminfo = cfg.gc.get('/item', parameters={'text': 'TCGA-A2-A0YE-01Z-00-DX1'})[0]

    cfg.folder = cfg.gc.post(
        '/folder', data={
            'parentId': original_iteminfo['folderId'],
            'name': 'test-annot-and-mask-handler',
        })

    # copy the item
    cfg.iteminfo = cfg.gc.post(
        '/item/%s/copy' % original_iteminfo['_id'], data={
            'name': 'TCGA-A2-A0YE-01Z.svs',
            'copyAnnotations': True,
            'folderId': cfg.folder['_id'],
        })

    # read GTCodes dataframe
    gtcodePath = getTestFilePath('sample_GTcodes.csv')
    cfg.GTcodes = read_csv(gtcodePath)
    cfg.GTcodes.index = cfg.GTcodes.loc[:, 'group']

    # other params
    cfg.get_roi_mask_kwargs = {
        'iou_thresh': 0.0,
        'crop_to_roi': True,
        'use_shapely': True,
        'verbose': False,
    }
    cfg.get_contours_kwargs = {
        'groups_to_get': None,
        'roi_group': 'roi',
        'get_roi_contour': True,
        'discard_nonenclosed_background': True,
        'background_group': 'mostly_stroma',
        'MIN_SIZE': 10, 'MAX_SIZE': None,
        'verbose': False, 'monitorPrefix': '',
    }

    # Microns-per-pixel / Magnification (either or)
    cfg.MPP = 5.0
    cfg.MAG = None

    # get annotations for slide
    cfg.slide_annotations = cfg.gc.get('/annotation/item/' + cfg.iteminfo['_id'])

    # scale up/down annotations by a factor
    sf, _ = get_scale_factor_and_appendStr(
        gc=cfg.gc, slide_id=cfg.iteminfo['_id'], MPP=cfg.MPP, MAG=cfg.MAG)
    cfg.slide_annotations = scale_slide_annotations(cfg.slide_annotations, sf=sf)

    # get bounding box information for all annotations
    cfg.element_infos = get_bboxes_from_slide_annotations(cfg.slide_annotations)

    # params for get_image_and_mask_from_slide()
    cfg.get_kwargs = {
        'gc': cfg.gc, 'slide_id': cfg.iteminfo['_id'],
        'GTCodes_dict': cfg.GTcodes.T.to_dict(),
        'bounds': {
            'XMIN': 58000, 'XMAX': 63000,
            'YMIN': 35000, 'YMAX': 39000},
        'MPP': cfg.MPP,
        'MAG': cfg.MAG,
        'get_roi_mask_kwargs': cfg.get_roi_mask_kwargs,
        'get_contours_kwargs': cfg.get_contours_kwargs,
        'get_rgb': True,
        'get_contours': True,
        'get_visualization': True,
    }


class TestGetROIMasks:
    """Test methods for getting ROI mask from annotations."""

    def test_get_roi_mask(self):
        """Test get_roi_mask()."""
        # get indices of rois
        idxs_for_all_rois = _get_idxs_for_all_rois(
            GTCodes=cfg.GTcodes, element_infos=cfg.element_infos.copy())

        # get roi mask and info
        ROI, roiinfo = get_roi_mask(
            slide_annotations=copy.deepcopy(cfg.slide_annotations),
            element_infos=cfg.element_infos.copy(),
            GTCodes_df=cfg.GTcodes.copy(),
            idx_for_roi=idxs_for_all_rois[0],  # <- let's focus on first ROI,
            roiinfo=None, **cfg.get_roi_mask_kwargs)

        assert ROI.shape == (228, 226)
        assert (
            roiinfo['BBOX_HEIGHT'], roiinfo['BBOX_WIDTH'],
            roiinfo['XMIN'], roiinfo['XMAX'],
            roiinfo['YMIN'], roiinfo['YMAX']) == (
            242, 351, 2966, 3317, 1678, 1920)

    def test_get_all_rois_from_slide(self, tmpdir):  # noqa
        """Test get_all_roi_masks_for_slide() as-is without tiling."""
        # just a temp directory to save masks for now
        base_savepath = str(tmpdir)
        savepaths = {
            'ROI': os.path.join(base_savepath, 'masks'),
            'rgb': os.path.join(base_savepath, 'rgbs'),
            'contours': os.path.join(base_savepath, 'contours'),
            'visualization': os.path.join(base_savepath, 'vis'),
        }
        for _, savepath in savepaths.items():
            os.mkdir(savepath)

        detailed_kwargs = {
            'MPP': cfg.MPP,
            'MAG': None,
            'get_roi_mask_kwargs': cfg.get_roi_mask_kwargs,
            'get_contours_kwargs': cfg.get_contours_kwargs,
            'get_rgb': True,
            'get_contours': True,
            'get_visualization': True,
        }

        savenames = get_all_rois_from_slide(
            gc=cfg.gc, slide_id=cfg.iteminfo['_id'],
            GTCodes_dict=cfg.GTcodes.T.to_dict(), save_directories=savepaths,
            get_image_and_mask_from_slide_kwargs=detailed_kwargs,
            max_roiside=None,
            slide_name='TCGA-A2-A0YE', verbose=False)

        assert len(savenames) == 3
        assert set(savenames[0].keys()) == {
            'ROI', 'rgb', 'visualization', 'contours'}
        assert {
            'TCGA-A2-A0YE_left-57604_top-35808_bottom-37445_right-59441.png',
            'TCGA-A2-A0YE_left-58483_top-38223_bottom-39780_right-60399.png',
            'TCGA-A2-A0YE_left-59201_top-33493_bottom-38063_right-63732.png',
        } == {os.path.basename(savename['ROI']) for savename in savenames}

    def test_get_all_rois_from_slide_tiled(self, tmpdir):  # noqa
        """Test get_all_roi_masks_for_slide() with tiling."""
        # just a temp directory to save masks for now
        base_savepath = str(tmpdir)
        savepaths = {
            'ROI': os.path.join(base_savepath, 'masks'),
            'rgb': os.path.join(base_savepath, 'rgbs'),
            'contours': os.path.join(base_savepath, 'contours'),
            'visualization': os.path.join(base_savepath, 'vis'),
        }
        for _, savepath in savepaths.items():
            os.mkdir(savepath)

        detailed_kwargs = {
            'MPP': cfg.MPP,
            'MAG': None,
            'get_roi_mask_kwargs': cfg.get_roi_mask_kwargs,
            'get_contours_kwargs': None,
            'get_rgb': False,
            'get_contours': False,
            'get_visualization': False,
        }

        savenames = get_all_rois_from_slide(
            gc=cfg.gc, slide_id=cfg.iteminfo['_id'],
            GTCodes_dict=cfg.GTcodes.T.to_dict(), save_directories=savepaths,
            get_image_and_mask_from_slide_kwargs=detailed_kwargs,
            max_roiside=128,
            slide_name='TCGA-A2-A0YE', verbose=False)

        assert len(savenames) == 6
        assert set(savenames[0].keys()) == {'ROI'}
        assert {
            'TCGA-A2-A0YE_left-59201_top-33493_bottom-36047_right-61756.png',
            'TCGA-A2-A0YE_left-59201_top-36047_bottom-38063_right-61756.png',
            'TCGA-A2-A0YE_left-61756_top-33493_bottom-36047_right-63732.png',
            'TCGA-A2-A0YE_left-61756_top-36047_bottom-38063_right-63732.png',
            'TCGA-A2-A0YE_left-58483_top-38223_bottom-39780_right-60399.png',
            'TCGA-A2-A0YE_left-57604_top-35808_bottom-37445_right-59441.png',
        } == {os.path.basename(savename['ROI']) for savename in savenames}

    def test_get_image_and_mask_manual_bounds(self):
        """Test get_image_and_mask_from_slide()."""
        # get specified region -- without providing scaled annotations
        roi_out_1 = get_image_and_mask_from_slide(
            mode='manual_bounds', **cfg.get_kwargs)

        # get specified region -- with providing scaled annotations
        roi_out_2 = get_image_and_mask_from_slide(
            mode='manual_bounds',
            slide_annotations=copy.deepcopy(cfg.slide_annotations),
            element_infos=cfg.element_infos.copy(), **cfg.get_kwargs)

        for roi_out in (roi_out_1, roi_out_2):
            assert set(roi_out.keys()) == {
                'bounds', 'ROI', 'rgb', 'contours', 'visualization'}
            assert roi_out['ROI'].shape == (200, 250)
            assert roi_out['rgb'].shape == (200, 250, 3)
            assert roi_out['visualization'].shape == (200, 250, 3)
            assert len(roi_out['contours']) > 26
            assert (
                len(roi_out['contours']) < 32)
            assert set(roi_out['contours'][0].keys()) == {
                'group', 'color', 'ymin', 'ymax', 'xmin', 'xmax',
                'has_holes', 'touches_edge-top', 'touches_edge-left',
                'touches_edge-bottom', 'touches_edge-right', 'coords_x',
                'coords_y',
            }

    def test_get_image_and_mask_minbbox(self):
        """Test get_image_and_mask_from_slide()."""
        # get ROI bounding everything
        roi_out = get_image_and_mask_from_slide(
            mode='min_bounding_box',
            slide_annotations=copy.deepcopy(cfg.slide_annotations),
            element_infos=cfg.element_infos.copy(), **cfg.get_kwargs)

        assert set(roi_out.keys()) == {
            'bounds', 'ROI', 'rgb', 'contours', 'visualization'}
        assert roi_out['ROI'].shape == (321, 351)
        assert roi_out['rgb'].shape == (321, 351, 3)
        assert roi_out['visualization'].shape == (321, 351, 3)
        assert len(roi_out['contours']) > 26
        assert (
            len(roi_out['contours']) < 32)
        assert set(roi_out['contours'][0].keys()) == {
            'group', 'color', 'ymin', 'ymax', 'xmin', 'xmax',
            'has_holes', 'touches_edge-top', 'touches_edge-left',
            'touches_edge-bottom', 'touches_edge-right', 'coords_x',
            'coords_y',
        }



import numpy as np


def condense(im_label):
    """
    Shifts labels in a label image to fill in gaps corresponding to missing
    values.

    Parameters
    ----------
    im_label : array_like
        A label image generated by segmentation methods.

    Returns
    -------
    Condensed : array_like
        A label image where all values > 0 are shifted down to fill gaps.

    See Also
    --------
    histomicstk.segmentation.label.shuffle

    """
    import scipy.ndimage as ndi

    # initialize output
    Condensed = im_label.copy()

    # get extent of each object
    Locations = ndi.find_objects(Condensed)

    # initialize counter
    Counter = 1

    # fill in new values
    for i in np.arange(1, len(Locations) + 1):
        if Locations[i - 1] is not None:
            Patch = Condensed[Locations[i - 1]]
            Patch[Patch == i] = Counter
            Counter += 1

    return Condensed



import json
from pathlib import Path

import histomicstk
from histomicstk.cli import utils
from histomicstk.cli.utils import CLIArgumentParser
from histomicstk.preprocessing.color_normalization import background_intensity


def main(args):
    other_args = {'outputAnnotationFile', 'scheduler'}
    kwargs = {k: v for k, v in vars(args).items()
              if k not in other_args}
    # Allow (some) default parameters to work.  Assume certain values
    # are not valid.
    for k in 'sample_fraction', 'sample_approximate_total':
        if kwargs[k] == -1:
            del kwargs[k]

    utils.create_dask_client(args)
    I_0 = background_intensity(**kwargs)

    #
    # Write annotation file
    #

    print('\n>> Writing annotation file ...\n')

    annotation = {
        'name': 'BackgroundIntensity',
        'attributes': {
            'intensity_values': I_0.tolist(),
            'params': vars(args),
            'cli': Path(__file__).stem,
            'version': histomicstk.__version__,
        },
    }

    with open(args.outputAnnotationFile, 'w') as annotation_file:
        json.dump(annotation, annotation_file, separators=(',', ':'), sort_keys=False)


if __name__ == '__main__':
    main(CLIArgumentParser().parse_args())



import numpy as np


def clog(im_input, im_mask, sigma_min, sigma_max):
    """Constrained Laplacian of Gaussian filter.

    Takes as input a grayscale nuclear image and binary mask of cell nuclei,
    and uses the distance transform of the nuclear mask to constrain the LoG
    filter response of the image for nuclear seeding. Returns a LoG filter
    image of type float. Local maxima are used for seeding cells.

    Parameters
    ----------
    im_input : array_like
        A hematoxylin intensity image obtained from ColorDeconvolution. Objects
        are assumed to be dark with a light background.
    im_mask : array_like
        A binary image where nuclei pixels have value 1/True, and non-nuclear
        pixels have value 0/False.
    sigma_min : double
        Minimum sigma value for the scale space. For blob detection, set this
        equal to minimum-blob-radius / sqrt(2).
    sigma_max : double
        Maximum sigma value for the scale space. For blob detection, set this
        equal to maximum-blob-radius / sqrt(2).

    Returns
    -------
    im_log_max : array_like
        An intensity image containing the maximal LoG filter response across
        all scales for each pixel
    im_sigma_max : array_like
        An intensity image containing the sigma value corresponding to the
        maximal LoG response at each pixel. The nuclei/blob radius value for
        a given sigma can be estimated to be equal to sigma * sqrt(2).

    References
    ----------
    .. [#] Y. Al-Kofahi et al "Improved Automatic Detection and Segmentation
           of Cell Nuclei in Histopathology Images" in IEEE Transactions on
           Biomedical Engineering,vol.57,no.4,pp.847-52, 2010.

    """
    from scipy.ndimage import distance_transform_edt, gaussian_laplace

    # convert intensity image type to float
    im_input = im_input.astype(float)

    # generate distance map
    im_dmap = distance_transform_edt(im_mask)

    # compute max sigma at each pixel as 2 times the distance to background
    im_sigma_ubound = 2.0 * im_dmap

    # clip max sigma values to specified range
    im_sigma_ubound = np.clip(im_sigma_ubound, sigma_min, sigma_max)

    # initialize log filter response array
    MIN_FLOAT = np.finfo(im_input.dtype).min

    im_log_max = np.zeros_like(im_input)
    im_log_max[:, :] = MIN_FLOAT

    im_sigma_max = np.zeros_like(im_input)

    # Compute maximal LoG filter response across the scale space
    sigma_start = int(np.floor(sigma_min))
    sigma_end = int(np.ceil(sigma_max))

    sigma_list = np.linspace(sigma_start, sigma_end,
                             int(sigma_end - sigma_start + 1))

    for sigma in sigma_list:

        # generate normalized filter response
        im_log_cur = sigma**2 * gaussian_laplace(im_input, sigma, mode='mirror')

        # constrain LoG response
        im_log_cur[im_sigma_ubound < sigma] = MIN_FLOAT

        # update maxima
        max_update_pixels = np.where(im_log_cur > im_log_max)

        if len(max_update_pixels[0]) > 0:

            im_log_max[max_update_pixels] = im_log_cur[max_update_pixels]
            im_sigma_max[max_update_pixels] = sigma

    # replace min floats
    im_log_max[im_log_max == MIN_FLOAT] = 0

    return im_log_max, im_sigma_max



from collections import namedtuple

import large_image
import numpy as np

from ..preprocessing.color_conversion import rgb_to_hsi


# This can be an enum in Python >= 3.4
class Labels:
    """Labels for the output image of the positive pixel count routines."""

    NEGATIVE = 0
    WEAK = 1
    PLAIN = 2
    STRONG = 3


class Parameters(
        namedtuple('Parameters', [
            'hue_value',
            'hue_width',
            'saturation_minimum',
            'intensity_upper_limit',
            'intensity_weak_threshold',
            'intensity_strong_threshold',
            'intensity_lower_limit',
        ]),
):
    """Parameters(hue_value, hue_width, saturation_minimum,
    intensity_upper_limit, intensity_weak_threshold,
    intensity_strong_threshold, intensity_lower_limit)

    Attributes
    ----------
    hue_value:
      Center of the hue range in HSI space for the positive color, in
      the range [0, 1]
    hue_width:
      Width of the hue range in HSI space
    saturation_minimum:
      Minimum saturation of positive pixels in HSI space, in the range
      [0, 1]
    intensity_upper_limit:
      Intensity threshold in HSI space above which a pixel is
      considered negative, in the range [0, 1]
    intensity_weak_threshold:
      Intensity threshold in HSI space that separates weak-positive
      pixels (above) from plain positive pixels (below)
    intensity_strong_threshold:
      Intensity threshold in HSI space that separates plain positive
      pixels (above) from strong positive pixels (below)
    intensity_lower_limit:
      Intensity threshold in HSI space below which a pixel is
      considered negative

    """


OutputTotals = namedtuple('OutputTotals', [
    'NumberWeakPositive',
    'NumberPositive',
    'NumberStrongPositive',
    'NumberTotalPixels',
    'IntensitySumWeakPositive',
    'IntensitySumPositive',
    'IntensitySumStrongPositive',
])

Output = namedtuple('Output', OutputTotals._fields + (
    'IntensityAverage',
    'RatioStrongToTotal',
    'IntensityAverageWeakAndPositive',
    'RatioStrongToPixels',
    'RatioWeakToPixels',
    'RatioTotalToPixels',
))


def count_slide(slide_path, params, region=None,
                tile_grouping=256, make_label_image=False):
    """Compute a count of positive pixels in the slide at slide_path.
    This routine can also create a label image.

    Parameters
    ----------
    slide_path : string (path)
        Path to the slide to analyze.
    params : Parameters
        An instance of Parameters, which see for further documentation
    region : dict, optional
        A valid region dict (per a large_image
        TileSource.tileIterator's region argument)
    tile_grouping : int
        The number of tiles to process as part of a single task
    make_label_image : bool, default=False
        Whether to make a label image.  See also "Notes"

    Returns
    -------
    stats : Output
        Various statistics on the input image.  See Output.
    label_image : array-like, only if make_label_image is set

    Notes
    -----
    The return value is either a single or a pair -- it is in either
    case a tuple.  Dask is used as configured to compute the
    statistics, but only if make_label_image is reset.  If
    make_label_image is set, everything is computed in a
    single-threaded manner.

    """
    from dask import delayed

    ts = large_image.getTileSource(slide_path)
    kwargs = dict(format=large_image.tilesource.TILE_FORMAT_NUMPY)
    if region is not None:
        kwargs['region'] = region
    if make_label_image:
        tile = ts.getRegion(**kwargs)[0]
        return count_image(tile, params)
    else:
        results = []
        total_tiles = ts.getSingleTile(**kwargs)['iterator_range']['position']
        for position in range(0, total_tiles, tile_grouping):
            results.append(delayed(_count_tiles)(
                slide_path, params, kwargs, position,
                min(tile_grouping, total_tiles - position)))
        results = delayed(_combine)(results).compute()
    return (_totals_to_stats(results),)


def _combine(results):
    return OutputTotals._make(sum(r[i] for r in results)
                              for i in range(len(OutputTotals._fields)))


def _count_tiles(slide_path, params, kwargs, position, count):
    ts = large_image.getTileSource(slide_path)
    lpotf = len(OutputTotals._fields)
    total = [0] * lpotf
    for pos in range(position, position + count):
        tile = ts.getSingleTile(tile_position=pos, **kwargs)['tile']
        subtotal = _count_image(tile, params)[0]
        for k in range(lpotf):
            total[k] += subtotal[k]
    return OutputTotals._make(total)


def count_image(image, params, mask=None):
    """Count positive pixels, computing a label mask and summary
    statistics.

    Parameters
    ----------
    image : array-like
        NxMx3 array of RGB data
    params : Parameters
        An instance of Parameters, which see for further documentation
    mask: array-like
        A boolean mask.  If present, only pixels where the mask is True are
        considered.

    Returns
    -------
    stats : Output
        Various statistics on the input image.  See Output.
    label_image : array-like
        NxM array of pixel types.  See Labels for the different values.

    """
    total, masks = _count_image(image, params, mask)
    mask_all_positive, mask_weak, mask_pos, mask_strong = masks
    label_image = np.full(image.shape[:-1], Labels.NEGATIVE, dtype=np.uint8)
    label_image[mask_all_positive] = (
        mask_weak * Labels.WEAK +
        mask_pos * Labels.PLAIN +
        mask_strong * Labels.STRONG
    )
    return _totals_to_stats(total), label_image


def _count_image(image, params, mask=None):
    """A version of count_image that doesn't compute the label image and
    only computes the sums.

    """
    p = params
    image_hsi = rgb_to_hsi(image / 255)
    mask_all_positive = (
        (np.abs(((image_hsi[..., 0] - p.hue_value + 0.5) % 1) - 0.5) <=
         p.hue_width / 2) &
        (image_hsi[..., 1] >= p.saturation_minimum) &
        (image_hsi[..., 2] < p.intensity_upper_limit) &
        (image_hsi[..., 2] >= p.intensity_lower_limit)
    )
    if mask is not None:
        mask_all_positive &= mask
    all_positive_i = image_hsi[mask_all_positive, 2]
    mask_weak = all_positive_i >= p.intensity_weak_threshold
    nw, iw = np.count_nonzero(mask_weak), np.sum(all_positive_i[mask_weak])
    mask_strong = all_positive_i < p.intensity_strong_threshold
    ns, is_ = np.count_nonzero(mask_strong), np.sum(all_positive_i[mask_strong])
    mask_pos = ~(mask_weak | mask_strong)
    np_, ip = np.count_nonzero(mask_pos), np.sum(all_positive_i[mask_pos])
    total = OutputTotals(
        NumberWeakPositive=nw,
        NumberPositive=np_,
        NumberStrongPositive=ns,
        NumberTotalPixels=(np.count_nonzero(mask) if mask is not None else
                           image_hsi.shape[0] * image_hsi.shape[1]),
        IntensitySumWeakPositive=iw,
        IntensitySumPositive=ip,
        IntensitySumStrongPositive=is_,
    )
    return total, (mask_all_positive, mask_weak, mask_pos, mask_strong)


def _totals_to_stats(total):
    """Do the extra computations to convert an OutputTotals to an Output"""
    t = total
    all_positive = t.NumberWeakPositive + t.NumberPositive + t.NumberStrongPositive
    return Output(
        IntensityAverage=((t.IntensitySumWeakPositive +
                           t.IntensitySumPositive +
                           t.IntensitySumStrongPositive) /
                          all_positive) if all_positive else 0,
        RatioStrongToTotal=t.NumberStrongPositive / all_positive if all_positive else 0,
        IntensityAverageWeakAndPositive=(
            (t.IntensitySumWeakPositive + t.IntensitySumPositive) /
            (t.NumberWeakPositive + t.NumberPositive)
        ) if t.NumberWeakPositive + t.NumberPositive else 0,
        RatioStrongToPixels=t.NumberStrongPositive / (t.NumberTotalPixels or 1),
        RatioWeakToPixels=t.NumberWeakPositive / (t.NumberTotalPixels or 1),
        RatioTotalToPixels=all_positive / (t.NumberTotalPixels or 1),
        **t._asdict(),
    )


__all__ = (
    'Parameters',
    'Output',
    'count_slide',
    'count_image',
)



"""
Created on Mon Sep 30 18:12:48 2019.

@author: mtageld
"""
import os
import sys

import pytest

from histomicstk.utils.girder_convenience_utils import \
    update_styles_for_annotations_in_folder
from histomicstk.workflows.workflow_runner import Slide_iterator

thisDir = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, os.path.join(thisDir, '../../../'))
from tests.htk_test_utilities import girderClient  # noqa


class Cfg:
    def __init__(self):
        self.gc = None
        self.posted_folder = None


cfg = Cfg()


class TestWorkflows:
    """Test slide and folder iterator & workflow runner."""

    # pytest runs tests in the order they appear in the module
    @pytest.mark.usefixtures('girderClient')  # noqa
    def test_prep(self, girderClient):  # noqa

        cfg.gc = girderClient

        # get original item
        original_iteminfo = cfg.gc.get('/item', parameters={'text': 'TCGA-A2-A0YE-01Z-00-DX1'})[0]

        # create a sample folder
        cfg.posted_folder = cfg.gc.post(
            '/folder', data={
                'parentId': original_iteminfo['folderId'],
                'name': 'test-workflow',
            })

        # copy the item a couple of times
        for i in range(2):
            _ = cfg.gc.post(
                '/item/%s/copy' % original_iteminfo['_id'], data={
                    'name': 'TCGA-A2-A0YE_copy-%d' % i,
                    'copyAnnotations': True,
                    'folderId': cfg.posted_folder['_id'],
                })

    def test_Slide_iterator(self):
        """Test Slide_iterator.run()."""
        slide_iterator = Slide_iterator(
            cfg.gc, source_folder_id=cfg.posted_folder['_id'])

        assert len(slide_iterator.slide_ids) == 2

        si = slide_iterator.run()
        for i in range(2):
            slide_info = next(si)
            if i == 0:
                assert all(
                    k in slide_info.keys() for k in
                    ('name', '_id', 'levels', 'magnification'))

    def test_runner_using_annotation_style_update(self):
        """Test workflow runner for cellularity detection."""
        update_styles_for_annotations_in_folder(
            cfg.gc, folderid=cfg.posted_folder['_id'],
            workflow_kwargs={
                'changes': {
                    'roi': {
                        'group': 'modified_roi',
                        'lineColor': 'rgb(0,0,255)',
                        'fillColor': 'rgba(0,0,255,0.3)',
                    },
                },
            },
            recursive=True, catch_exceptions=True,
            monitor='test', verbose=0,
        )
        # just make sure it runs without exceptions and yo're good.



import numpy as np

from .condense import condense
from .delete import delete


def delete_border(im_label):
    """
    Deletes objects touching the border of the image and relabel

    Parameters
    ----------
    im_label : array_like
        A label mask generated by segmentation methods.

    Returns
    -------
    im_label_del : array_like
        A label mask where in all the objects touching the image border have
        been deleted

    See Also
    --------
    histomicstk.segmentation.label.delete

    """
    if not np.any(im_label):
        return im_label

    im_border_mask = np.zeros_like(im_label)
    im_border_mask[:, 0] = 1
    im_border_mask[:, -1] = 1
    im_border_mask[0, :] = 1
    im_border_mask[-1, :] = 1

    border_indices = np.unique(im_label[im_border_mask > 0])
    border_indices = border_indices[border_indices > 0]

    if len(border_indices) == 0:
        return im_label

    # Condense and delete the border regions from the image label
    im_label_del = np.zeros_like(im_label)
    im_label_del = condense(delete(im_label, border_indices))
    return im_label_del



import numpy as np


def shuffle(im_label):
    """
    Shuffles labels in a label image to improve visualization and enhance
    object boundaries.

    Parameters
    ----------
    im_label : array_like
        A label image generated by segmentation methods.

    Returns
    -------
    Shuffled : array_like
        A label image where all values > 0 are randomly shuffled.

    See Also
    --------
    histomicstk.segmentation.label.CondenseLabel

    """
    from skimage import measure as ms

    # get list of unique object labels
    Unique = np.unique(im_label.flatten())

    # remove background objects (im_label == 0)
    Unique = np.delete(Unique, (Unique == 0).nonzero())

    # generate shuffled list of object values
    np.random.shuffle(Unique)

    # initialize output
    Shuffled = np.zeros(im_label.shape, dtype=np.uint32)

    # get pixel list for each object
    Props = ms.regionprops(im_label.astype(int))

    # fill in new values
    for i in range(len(Unique)):
        Coords = Props[i].coords
        Shuffled[Coords[:, 0], Coords[:, 1]] = Unique[i]

    return Shuffled



"""
Created on Sun Oct 20 00:14:03 2019.

@author: mtageld
"""
import os
import sys

import numpy as np
from skimage.transform import resize

from histomicstk.preprocessing.augmentation.color_augmentation import \
    rgb_perturb_stain_concentration
from histomicstk.preprocessing.color_conversion import lab_mean_std
from histomicstk.preprocessing.color_normalization import reinhard
from histomicstk.preprocessing.color_normalization.deconvolution_based_normalization import \
    deconvolution_based_normalization
from histomicstk.saliency.tissue_detection import get_tissue_mask

thisDir = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, os.path.join(thisDir, '../../../tests'))
# import htk_test_utilities as utilities  # noqa
from datastore import datastore  # noqa
from htk_test_utilities import getTestFilePath, girderClient  # noqa


class Cfg:
    def __init__(self):
        self.gc = None
        self.tissue_rgb = None
        self.mask_out = None


cfg = Cfg()


def test_prep():

    import large_image

    src = datastore.fetch('TCGA-A2-A0YE-01Z-00-DX1.8A2E3094-5755-42BC-969D-7F0A2ECA0F39.svs')
    ts = large_image.open(src)
    MAG = 1.5
    cfg.tissue_rgb = ts.getRegion(
        region=dict(left=46890, right=50000, top=40350, bottom=43000),
        scale=dict(magnification=MAG),
        format='numpy')[0][:, :, :3]
    # get mask of things to ignore
    cfg.mask_out, _ = get_tissue_mask(
        cfg.tissue_rgb, deconvolve_first=False,
        n_thresholding_steps=1, sigma=1.5, min_size=30)
    cfg.mask_out = resize(
        cfg.mask_out == 0, output_shape=cfg.tissue_rgb.shape[:2],
        order=0, preserve_range=True) == 1


class TestColorNormalization():
    """Test color normalization."""

    def test_reinhard(self):
        """Test reinhard."""
        # # SANITY CHECK! normalize to LAB mean and std from SAME slide
        # mean_lab, std_lab = lab_mean_std(tissue_rgb)
        # tissue_rgb_normalized = reinhard(
        #     tissue_rgb, target_mu=mean_lab, target_sigma=std_lab)
        # # we expect the images to be (almost) exactly the same
        # assert np.mean(tissue_rgb - tissue_rgb_normalized) < 1

        # color norm. standard (from TCGA-A2-A3XS-DX1, Amgad et al, 2019)
        cnorm = {
            'mu': np.array([8.74108109, -0.12440419, 0.0444982]),
            'sigma': np.array([0.6135447, 0.10989545, 0.0286032]),
        }

        # Normalize to pre-set color standard (unmasked)
        tissue_rgb_normalized = reinhard(
            cfg.tissue_rgb, target_mu=cnorm['mu'], target_sigma=cnorm['sigma'])

        # check that it matches
        mean_lab, std_lab = lab_mean_std(tissue_rgb_normalized)
        assert all(np.abs(mean_lab - cnorm['mu']) < [0.1, 0.1, 0.1])
        assert all(np.abs(std_lab - cnorm['sigma']) < [0.1, 0.1, 0.1])

        # Do MASKED normalization to preset standard
        tissue_rgb_normalized = reinhard(
            cfg.tissue_rgb, target_mu=cnorm['mu'], target_sigma=cnorm['sigma'],
            mask_out=cfg.mask_out)

        # check that it matches
        mean_lab, std_lab = lab_mean_std(
            tissue_rgb_normalized, mask_out=cfg.mask_out)
        assert all(np.abs(mean_lab - cnorm['mu']) < [0.1, 0.1, 0.1])
        assert all(np.abs(std_lab - cnorm['sigma']) < [0.1, 0.1, 0.1])

    def test_macenko_normalization(self):
        """Test macenko_pca normalization."""
        stain_unmixing_routine_params = {
            'stains': ['hematoxylin', 'eosin'],
            'stain_unmixing_method': 'macenko_pca',
        }

        # TCGA-A2-A3XS-DX1_xmin21421_ymin37486_.png, Amgad et al, 2019)
        # for macenco (obtained using rgb_separate_stains_macenko_pca()
        # and using reordered such that columns are the order:
        # Hamtoxylin, Eosin, Null
        W_target = np.array([
            [0.5807549, 0.08314027, 0.08213795],
            [0.71681094, 0.90081588, 0.41999816],
            [0.38588316, 0.42616716, -0.90380025],
        ])

        # Macenko - Unmasked, using default, 'idealized' W_target"
        tissue_rgb_normalized = deconvolution_based_normalization(
            cfg.tissue_rgb,
            stain_unmixing_routine_params=stain_unmixing_routine_params)
        assert tuple(
            [int(tissue_rgb_normalized[..., i].mean()) for i in range(3)],
        ) in {(183, 121, 212), (186, 128, 215)}

        # Macenko - Unmasked, using W_target from good image
        tissue_rgb_normalized = deconvolution_based_normalization(
            cfg.tissue_rgb, W_target=W_target,
            stain_unmixing_routine_params=stain_unmixing_routine_params)
        assert tuple(
            [int(tissue_rgb_normalized[..., i].mean()) for i in range(3)],
        ) in {(188, 125, 175), (188, 124, 175), (191, 132, 179)}

        # Macenko - Masked, using W_target from good image
        tissue_rgb_normalized = deconvolution_based_normalization(
            cfg.tissue_rgb, W_target=W_target, mask_out=cfg.mask_out,
            stain_unmixing_routine_params=stain_unmixing_routine_params)
        assert tuple(
            [int(tissue_rgb_normalized[..., i].mean()) for i in range(3)],
        ) in {(188, 125, 175), (187, 125, 174), (192, 131, 179)}


class TestColorAugmentation:
    """Test color augmentation."""

    def test_rgb_perturb_stain_concentration(self):
        """Test rgb_perturb_stain_concentration."""
        # for reproducibility
        np.random.seed(0)

        # Unmasked
        augmented_rgb = rgb_perturb_stain_concentration(cfg.tissue_rgb)
        assert tuple(
            [int(augmented_rgb[..., i].mean()) for i in range(3)],
        ) in {(178, 115, 154), (177, 114, 153), (190, 116, 168)}

        # Masked
        augmented_rgb = rgb_perturb_stain_concentration(
            cfg.tissue_rgb, mask_out=cfg.mask_out)
        assert tuple(
            [int(augmented_rgb[..., i].mean()) for i in range(3)],
        ) in {(174, 101, 139), (174, 100, 139), (188, 102, 155)}
